_cosmo_labels = dict(
    cosmomc_theta=r"\theta_\mathrm{MC}",
    logA=r"\log(10^{10} A_\mathrm{s})",
    As=r"A_\mathrm{s}",
    Alens=r"A_\mathrm{lens}",
    ns=r"n_\mathrm{s}",
    ombh2=r"\Omega_\mathrm{b}h^2",
    omch2=r"\Omega_\mathrm{c}h^2",
    tau=r"\tau_\mathrm{reio}",
)
